# Copyright (C) 2002-2010
# Max-Planck-Institut fuer Radioastronomie Bonn
# Argelander Institut fuer Astronomie
# Astronomisches Institut der Ruhr-Universitaet Bochum
#
# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#


"""
NAM: BogliPlot.py (file)
DES: contains Bogli command handler class
"""
__version__ = '$Revision: 2722 $'
__date__ = '$Date: 2010-04-29 11:43:32 +0200 (jeu. 29 avril 2010) $'

# ----------------------------------------------------------------------------------
# ----- Import ---------------------------------------------------------------------
# -------------------------------------------------------------------------

from ppgplot import *
import numpy as np
import copy
import cPickle
import string
from rea import ReaMessageHandler
from rea.Utilities import solvePoly, Timing

from .BogliConfig import *
from . import DeviceHandler
from rea.fortran.fUtilities import replacenan
from rea.fortran.fStat import minmax

# Keep in mind if a wedge was plotted or not (used to overplot on
# image) or that the aspect keyword was used

wasWedge = 0
wasAspect = 0

# The number of plot to do (used to check for overplot)
numPlot = 1

# Used for maps :
transformationMatrix = np.array([0, 1, 0, 0, 0, 1], np.float)

# Add a dedicated message handler as attribute
__MessHand = ReaMessageHandler.MessHand('Plot')


# --------------------------------------------------------------------------------
# ----- methods ------------------------------------------------------------------
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
def erase():
    """
    DES: Erase any existing graphics.
    """
    pgeras()

# -------------------------------------------------------------------------


def clear():
    """
    DES: clear plot
    """

    global __MessHand, xAxis, yAxis, wasWedge, wasAspect

    # Do we really need this ? Couldn't erase be enough ?
    current_device = pgqid()
    __MessHand.debug("clearing plot of device " + repr(current_device))
    pgeras()

    # Resize the graph so that the resize command is not needed anymore only
    # for /xwin
    if not pgqinf('dev/type'):
        pgpage()

    # reinitialise log-scale flags - FS050515
    # TODO Do we need that ? if yes all the other attribute should be also set
    xAxis['log'] = 0
    yAxis['log'] = 0

    # reinitialise the wedge and aspect memory
    wasWedge = 0
    wasAspect = 0

# -------------------------------------------------------------------------


def setViewPoint(wedge=0, aspect=0):
    """
    DES: set the  view point
    OPT: (logical) wegde  : if wedge is present (default no)
         (logical) aspect : should we keep the aspect ratio (in 'physical' unit)
                            of the graph (default no)
    """

    global wasWedge, wasAspect, globalViewPoint, xAxis, yAxis

    # set the viewPoint
    if wedge == 1 or wasWedge == 1:
        # If wedge is present reserve space for it
        pgsvp(globalViewPoint['marging_left'],
              globalViewPoint['marging_right'] - globalViewPoint['wedge_size'],
              globalViewPoint['marging_bottom'], globalViewPoint['marging_top'])
    else:
        pgsvp(
            globalViewPoint['marging_left'], globalViewPoint['marging_right'],
            globalViewPoint['marging_bottom'], globalViewPoint['marging_top'])

    if aspect == 0 and wasAspect == 0:
        # and set the limits for the plots
        pgswin(xAxis['limits'][0], xAxis['limits'][1],
               yAxis['limits'][0], yAxis['limits'][1])
    else:
        pgwnad(xAxis['limits'][0], xAxis['limits'][1],
               yAxis['limits'][0], yAxis['limits'][1])

# -------------------------------------------------------------------------


def setLabels(labelX, labelY, caption):
    """
    DES: check and set labels
    INP: (strings) labelX, labelX, caption : the X and Y labels and the caption
    """

    global xAxis, yAxis, cLabel

    if isinstance(labelX, str):
        xAxis['text'] = labelX
    else:
        xAxis['text'] = 'x'
        __MessHand.warning("invalid x-label: " + repr(labelX))

    if isinstance(labelY, str):
        yAxis['text'] = labelY
    else:
        yAxis['text'] = 'y'
        __MessHand.warning("invalid y-label: " + repr(labelY))

    if isinstance(caption, str):
        cLabel['text'] = caption
    else:
        cLabel['text'] = ''
        __MessHand.warning("invalid caption: " + repr(caption))

# -------------------------------------------------------------------------


def labelling(wedge=0):
    """
    DES: Label x, y, caption and channel number.
    OPT: (logical) wedge : should we draw a wedge ? (default no)
    """

    global xAxis, yAxis, cLabel

    if xAxis['draw_it']:
        drawLabel(xAxis)
    if yAxis['draw_it']:
        drawLabel(yAxis)
    if cLabel['draw_it']:
        drawLabel(cLabel)

    if wedge:
        drawWedge()

# -------------------------------------------------------------------------


def drawWedge():
    """
    DES: Draw a wedge.
    """

    global zAxis

    # Maintained as a separate function but could be merged with
    # BogliPlot.Plot.labelling()...
    # TODO : think about that and fix

    if zAxis['draw_it']:

        pgsch(zAxis['charheight'])  # set character height
        pgsci(zAxis['color'])      # set colour index
        pgsls(zAxis['linestyle'])  # set line style
        pgslw(zAxis['linewidth'])  # set line width
        pgwedg(zAxis['side'],
               zAxis['displacement'],
               zAxis['width'],
               zAxis['limits'][1],
               zAxis['limits'][0],
               zAxis['text'])

# -------------------------------------------------------------------------


def drawLabel(label):
    """
    DES: generic function to draw labels/captions
    INP: (dict) label : a 'label' attribute like labelX
    """

    pgsch(label['charheight'])  # set character height
    pgsci(label['color'])      # set colour index
    pgmtxt(label['side'],
           label['displacement'],
           label['coordinate'],
           label['justification'],
           label['text'])

# -------------------------------------------------------------------------


def xyout(X, Y, text, size=0):
    """
    DES: generic function to overplot text
    INP: (int) X,Y : pixel position
         (str) text: text to output
         (flt) size: charheight (optional, default = use BogliConfig)
    """

    if size:
        pgsch(size)
    else:
        pgsch(xyouttext['charheight'])  # set character height
    pgsci(xyouttext['color'])      # set colour index
    pgptxt(X, Y,
           xyouttext['angle'],
           xyouttext['justification'],
           text)


# -------------------------------------------------------------------------
def plotDataXY(dataX, dataY, style='p', ci=1, width=0, ls=1):
    """
    DES: Plot x y data.
    INP: (array) dataX/Y : the array to be plotted (same dimension)
    OPT: (string) style : the style used for the plot ('l': line,
                          'p': point (default), 'b': histogram)
         (int)    ci    : the color index (default 1)
         (int)   width  : linewidth (def. 0 = use previous)
         (int)    ls    : linestyle (def. 1 = solid line)
    """

    myTiming = Timing()

    global xAxis, yAxis, point, line

    if xAxis['log']:
        dataX = np.log(np.array(dataX)) / np.log(10.)
    else:
        dataX = np.array(dataX)
    if yAxis['log']:
        dataY = np.log(np.array(dataY)) / np.log(10.)
    else:
        dataY = np.array(dataY)

    pgsci(ci)                                # set colour index

    if style == 'p':                       # points
        pgsch(point['size'])              # set character height
        pgpt(dataX, dataY, point['symbol'])  # plot points

    elif style == 'l':                       # lines
        if ls:
            pgsls(ls)
        else:
            pgsls(line['linestyle'])          # set line style
        if width:
            pgslw(width)                         # set line width
        else:
            pgslw(line['linewidth'])        # set line width
        pgline(dataX, dataY)                    # plot lines

    elif style == 'b':                       # bins
        pgsls(line['linestyle'])          # set line style
        pgslw(line['linewidth'])          # set line width
        pgbin(dataX, dataY)

    __MessHand.debug("plotDataXY done in " + str(myTiming))
# ----------------------------------------------------------------------------


def plotBox():
    """
    DES: Draw box and labels.
    """

    myTiming = Timing()

    global box, xAxis, yAxis, numPlot

    pgsch(box['charheight'])           # set character height
    pgsci(box['color'])                # set colour index
    pgsls(box['linestyle'])            # set line style
    pgslw(box['linewidth'])            # set line width

    if (xAxis['log']):
        if (yAxis['log']):
            pgbox('BCTSNL', 0.0, 0, 'BCTSNVL', 0.0, 0)
        else:
            pgbox('BCTSNL', 0.0, 0, 'BCTSNV', 0.0, 0)
    else:
        if (yAxis['log']):
            pgbox('BCTSN', 0.0, 0, 'BCTSNVL', 0.0, 0)
        else:
            pgbox('BCTSN', 0.0, 0, 'BCTSNV', 0.0, 0)

    __MessHand.debug("plotBox done in " + str(myTiming))

# -------------------------------------------------------------------------


def setLimits(dataX, dataY, limitsX=[], limitsY=[]):
    """
    DES: compute and/or set the limits for the graph
    INP: (arrays)        dataX/Y : the array to be plotted
    OPT: (2elts array) limitsX/Y : limits to use in X/Y for the plot
    """

    myTiming = Timing()

    global xAxis, yAxis

    if limitsX != []:
        # limitsX is set so use it
        xAxis['limits'] = np.array(limitsX)
    elif limitsY != []:
        # limitsX is not set but limitsY is set so compute the extrema
        # for dataX in the range defined by limitsY

        X_insideY = np.take(dataX, np.nonzero(np.greater_equal(dataY, min(limitsY)) &
                                              np.greater_equal(max(limitsY), dataY))[0])

        xAxis['limits'] = minmax(X_insideY)

    else:
        # nothing is set, compute the extrema
        xAxis['limits'] = minmax(dataX)

    if xAxis['log']:
        xAxis['limits'] = np.log(xAxis['limits']) / np.log(10.)

    # Same for Y
    if limitsY != []:
        yAxis['limits'] = np.array(limitsY)
    elif limitsX != []:
        Y_insideX = np.take(dataY, np.nonzero(np.greater_equal(dataX, np.min(limitsX)) &
                                              np.greater(np.max(limitsX), dataX))[0])

        yAxis['limits'] = minmax(Y_insideX)
    else:
        yAxis['limits'] = minmax(dataY)

    if yAxis['log']:
        yAxis['limits'] = np.log(yAxis['limits']) / np.log(10.)

    # If a limit is infinite, set it to 'nan'
    if str(xAxis['limits'][0]) == '-inf' or str(xAxis['limits'][0]) == 'inf':
        xAxis['limits'][0] = float('nan')
    if str(xAxis['limits'][1]) == '-inf' or str(xAxis['limits'][1]) == 'inf':
        xAxis['limits'][1] = float('nan')
    if str(yAxis['limits'][0]) == '-inf' or str(yAxis['limits'][0]) == 'inf':
        yAxis['limits'][0] = float('nan')
    if str(yAxis['limits'][1]) == '-inf' or str(yAxis['limits'][1]) == 'inf':
        yAxis['limits'][1] = float('nan')

    # Sometimes... an error occurs... and you ask for blank or
    # constant plots so...

    if xAxis['limits'][0] == xAxis['limits'][1]:
        xAxis['limits'][0] = xAxis['limits'][0] - \
            0.5  # this is completly arbitrary... but
        xAxis['limits'][1] = xAxis['limits'][1] + \
            0.5  # we are plotting constant so...

    if yAxis['limits'][0] == yAxis['limits'][1]:
        yAxis['limits'][0] = yAxis['limits'][0] - 0.5
        yAxis['limits'][1] = yAxis['limits'][1] + 0.5

    # add a margin to clearify the plot
    dist = max(xAxis['limits']) - min(xAxis['limits'])
    # crude way to determine the order of the limits
    if str(dist) != str(float('nan')):
        if xAxis['limits'][0] < xAxis['limits'][1]:
            xAxis['limits'] = xAxis['limits'] + \
                np.array([-1., 1.]) * addFracLim * dist
        else:
            xAxis['limits'] = xAxis['limits'] - \
                np.array([-1., 1.]) * addFracLim * dist

    dist = max(yAxis['limits']) - min(yAxis['limits'])
    if str(dist) != str(float('nan')):
        if yAxis['limits'][0] < yAxis['limits'][1]:
            yAxis['limits'] = yAxis['limits'] + \
                np.array([-1., 1.]) * addFracLim * dist
        else:
            yAxis['limits'] = yAxis['limits'] - \
                np.array([-1., 1.]) * addFracLim * dist

    __MessHand.debug("setLimits done in " + str(myTiming))


# -------------------------------------------------------------------------
def setMapTransformation(
        map_array, sizeX=[0., 1.], sizeY=[0., 1.], WCS=[]):
    """
    DES: compute transformation matrix
    """
    global transformationMatrix

    if len(map_array.shape) == 2:
        nX = map_array.shape[0]
        nY = map_array.shape[1]
    elif len(map_array.shape) == 3:
        nX = map_array[0].shape[0]
        nY = map_array[0].shape[1]

    if WCS != []:
        if WCS['CUNIT1'] == 'arcsec':
            cosY = np.cos(np.radians(WCS['CRVAL2']/ 3600.))
        else:
            cosY = np.cos(np.radians(WCS['CRVAL2']))
        cdeltCos = WCS['CDELT1'] / cosY
        transformationMatrix[0] = WCS['CRVAL1'] - (cdeltCos) * (WCS['CRPIX1'])
        transformationMatrix[1] = cdeltCos
        transformationMatrix[2] = 0
        transformationMatrix[3] = WCS[
            'CRVAL2'] - WCS['CDELT2'] * (WCS['CRPIX2'])
        transformationMatrix[4] = 0
        transformationMatrix[5] = WCS['CDELT2']

    else:

        # Making the assumption that the sizeX/sizeY array are sorted..
        # It could be useful to plot things in reverse order...

        x_pixel_size = (sizeX[1] - sizeX[0]) / (nX - 1)
        y_pixel_size = (sizeY[1] - sizeY[0]) / (nY - 1)

        transformationMatrix[0] = sizeX[0] - x_pixel_size  # to be centered
        transformationMatrix[1] = x_pixel_size
        transformationMatrix[3] = sizeY[0] - y_pixel_size
        transformationMatrix[5] = y_pixel_size


# -------------------------------------------------------------------------
def setMapLimits(map_array, limitsX=[], limitsY=[], limitsZ=[]):
    """
    DES: compute and/or set the limits for the map
    INP: (2D array)    map_array : the map array
    OPT: (2elts array) limitsX/Y : limits to use in X/Y for the plot
         (2elts array)  limitsZ  : the plotted color range in unit of the map_array
    """

    myTiming = Timing()

    global xAxis, yAxis, zAxis, transformationMatrix

    # Compute the map limits in X and Y

    if len(map_array.shape) == 2:
        nX = map_array.shape[0]
        nY = map_array.shape[1]
    elif len(map_array.shape) == 3:
        nX = map_array[0].shape[0]
        nY = map_array[0].shape[1]

    # if the limits on X/Y are given use them, or use all the map
    if limitsX != []:
        xAxis['limits'] = limitsX
    else:
        xAxis['limits'] = transformationMatrix[0] + \
            (transformationMatrix[1]) * (np.array([0, nX]) + 0.5)

    if limitsY != []:
        yAxis['limits'] = limitsY
    else:
        yAxis['limits'] = transformationMatrix[3] + \
            (transformationMatrix[5]) * (np.array([0, nY]) + 0.5)

    # if limits on Z are given use them
    # else compute the minmax of the map, restricted to the limitsX/Y
    if limitsZ != []:
        zAxis['limits'] = limitsZ
    else:
        # compute pixel positions corresponding to limitsX/Y
        tm = transformationMatrix
        x1 = int((xAxis['limits'][0] - tm[0]) / tm[1] + 0.5)
        x2 = int((xAxis['limits'][1] - tm[0]) / tm[1] + 0.5)
        y1 = int((yAxis['limits'][0] - tm[3]) / tm[5] + 0.5)
        y2 = int((yAxis['limits'][1] - tm[3]) / tm[5] + 0.5)
        lmap_array = np.ravel(np.array(map_array[x1:x2, y1:y2]))
        zAxis['limits'] = minmax(lmap_array)

    __MessHand.debug("setMapLimits done in " + str(myTiming))

# -------------------------------------------------------------------------


def setImaCol(
    style='g2r',
    contrast=1,
    brightness=0.5,
    transferFunction=0,
        nan=0):
    """
    NAM: setImaCol (method)
    DES: Set image colours.
    OPT: (string) style    : the style (default 'g2r' : green to red)
                             also defined 'r2g' red to green, 'b2r' blue to red,
                             'r2b' red to blue, 'blue' or any lut file define in
                             the BogliConfig.lutDir variable
         (float)  contrast   : the contrast of the plot (default 1)
         (float)  brightness : the brightness of the plot (default 0.5)
         (int) transferFunction : the transfer funtion (linear/log/sqrt)
    """

    (iCiLo, iCiHi) = pgqcir()  # color index used for the images
    pgscir(iCiLo, iCiHi)       # set colour index range

    ncolor = 255
    ramp_intensity = np.arange(ncolor * 1.0) / ncolor  # set the color ramp

    if style is 'g2r':
        red = 1.0 - ramp_intensity
        green = ramp_intensity
        blue = ramp_intensity * 0.0
    elif style is 'r2g':
        red = ramp_intensity
        green = 1.0 - ramp_intensity
        blue = ramp_intensity * 0.0
    elif style is 'b2r':
        red = 1.0 - ramp_intensity
        green = ramp_intensity * 0.0
        blue = ramp_intensity
    elif style is 'r2b':
        red = ramp_intensity
        green = ramp_intensity * 0.0
        blue = 1.0 - ramp_intensity
    elif style is 'blue':
        red = ramp_intensity * 0.0
        green = ramp_intensity * 0.0
        blue = ramp_intensity
    elif style in ['grey', 'gray']:
        red = 1.0 - ramp_intensity
        green = 1.0 - ramp_intensity
        blue = 1.0 - ramp_intensity
    elif style is 'inverse':
        red = ramp_intensity
        green = ramp_intensity
        blue = ramp_intensity

    else:
        red, green, blue = readLut(style)
        ncolor = len(red)
        ramp_intensity = np.arange(
            ncolor * 1.0) / ncolor  # set the color ramp

    pgctab(ramp_intensity, red, green, blue, ncolor, contrast, brightness)

    if nan:
        # Blank the lowest value (background) I do not understand why...
        pgscr(iCiHi, 0, 0, 0)

    pgsitf(transferFunction)  # set image transfer function
                             # ITF = 0 : linear
                             # ITF = 1 : logarithmic
                             # ITF = 2 : square-root

# -------------------------------------------------------------------------


def readLut(lutFile):
    """
    NAM: readLut (method)
    DES: read a LUT file
    INP: (string) lutFile : the name of the input lut file
    """

    try:
        f = file(lutDir + lutFile + '.lut')
    except IOError:
        __MessHand.setMess(1, " E: could not open file %s" % (lutFile))
        return

    # read and process LUT file
    lines = f.readlines()
    f.close()
    red, green, blue = [], [], []

    for i in range(len(lines)):
        tmp = string.split(lines[i])
        if tmp[0] != '!':			# skip comments
            red.append(string.atof(tmp[0]))
            green.append(string.atof(tmp[1]))
            blue.append(string.atof(tmp[2]))

    # Lut are defined in the inverse order so
    red.reverse()
    green.reverse()
    blue.reverse()

    # we need array to output
    return np.array(red), np.array(green), np.array(blue)

# -------------------------------------------------------------------------


def plot(dataX, dataY=[], limitsX=[], limitsY=[], labelX='x', labelY='y',
         caption='', style='p', ci=1, width=0, ls=1, overplot=0, aspect=0,
         logX=0, logY=0, nodata=0, noerase=0):
    """
    DES: do a plot
    INP: (array)  dataX = values to plot along X
         (array)  dataY = values to plot along Y (optional - default:
                                       plot dataX vs. running number)
    OPT: (2elts array) limitsX/Y = limits to use in X/Y for the plot
         (string)      labelX  = x label (default 'x')
         (string)      labelY  = y label (default 'y')
         (string)      caption = the caption of the plot (default ' ')
         (char)        style   = the style used for the plot ('l': line,
                                 'p': point (default), 'b': histogram)
         (int)            ci   = color index (default 1)
         (int)            ls   = line style (def 1) FB240307
         (int)          width  = linewidth (defaut 0 = use previous)
         (logical)      aspect = keep the aspect ratio in 'physical' unit
         (logical)    overplot = are we overplotting ? (default no)
         (logical)      logX/Y = do we use log scale ? (default no)
         (logical)      nodata = do not plot the data
         (logical)     noerase = do not clear the window
    """

    myTiming = Timing()

    global xAxis, yAxis, wasWedge, wasAspect, numPlot

    if dataY != []:
        dataX = np.array(dataX)
        dataY = np.array(dataY)
        # Make sure that X and Y have the same length
        if len(dataX) > len(dataY):
            dataX = dataX[:len(dataY)]
        elif len(dataY) > len(dataX):
            dataY = dataY[:len(dataX)]

    else:
        dataY = dataX
        dataX = np.arange(len(dataY))

    # Remove all NaN from X and Y arrays
    maskX = np.isnan(dataX)
    maskY = np.isnan(dataY)
    mNan = np.bitwise_or(maskX, maskY)
    bad = np.where(mNan)[0]
    if len(bad):
        # at least one NaN somewhere
        __MessHand.warning("%i data points NaN were found" % (len(bad)))
        good = np.nonzero(np.equal(mNan, 0))[0]
        if len(good):
            dataX = np.take(dataX, good)
            dataY = np.take(dataY, good)
        else:
            dataX = np.array([0], 'f')
            dataY = np.array([0], 'f')

    if not overplot:
        # Check for an open device first
        if not DeviceHandler.CurrentDev:
            __MessHand.warning('no device open')
            DeviceHandler.openDev()
        else:
            # to overcome the closing-without-closing feature of apexCalibrator
            DeviceHandler.__checkReopen()

        if not noerase:
            clear()

        if logX:
            xAxis['log'] = 1
        if logY:
            yAxis['log'] = 1
        setLimits(dataX, dataY, limitsX=limitsX, limitsY=limitsY)
        setLabels(labelX=labelX, labelY=labelY, caption=caption)

        wasAspect = aspect
        numPlot = 1
        setViewPoint(aspect=aspect)

    if overplot:
        if numPlot != 1:
            __MessHand.error('cannot overplot single Plot on multiPlot')
            return

    if not nodata:
        plotDataXY(dataX, dataY, style, ci, width, ls)

    if not overplot:
        if box['draw_it']:
            plotBox()
        labelling()

    __MessHand.debug("plot done in " + str(myTiming))

# --------------------------------------------------------------------------------
# Utilitie for one-D plots
#


def getPixel(order=0):
    """
    DES: allow user to get pixel values using mouse
    INP: (int)  order = for polynomial interpolation
    """

    __MessHand.info(
        "Click left to get one pixel, right (or 'e' or 'x') to exit")
    again = 1
    x, y = 0, 0
    nbData = 0
    posX = []  # store positions to perform polynomial interpolation
    posY = []
    __MessHand.info(str("(%0.10s, %0.10s)\n" % (str("%10s" % xAxis['text']),
                                                str("%10s" % yAxis['text']))))
    while(again):
        x, y, char = pgband(7, 0, x, y)
        if (char in ['X', 'x', 'e', 'E']):
            again = 0
        else:
            if xAxis['log']:
                xValue = 10. ** x
            else:
                xValue = x
            if yAxis['log']:
                yValue = 10. ** y
            else:
                yValue = y
            # display position
            __MessHand.info(str("(%10.3f, %10.3f)" % (xValue, yValue)))
            if order:
                # if (nbData > order):  # remove first point
                #  old=posX.pop(0)
                #  old=posY.pop(0)
                posX.append(x)        # append new one
                posY.append(y)
                if nbData:            # if at least two points
                    coeff = solvePoly(min([order, nbData]), posX, posY)
                    out = str(
                        " Interpolate %1i last points: Y = " %
                        (order + 1))
                    for n in range(order + 1):
                        if (coeff[n] > 0):
                            out += str("%+5.2f X^%1i " % (coeff[n], order - n))
                        else:
                            out += str("%6.2f X^%1i " % (coeff[n], order - n))
                    __MessHand.info(out)

                    # TODO: overplot the polynom
                nbData += 1


def draw(map_array, sizeX=[], sizeY=[], WCS=[],
         limitsX=[], limitsY=[], limitsZ=[],
         nan=1,
         labelX='x', labelY='y', caption='',
         style='g2r', contrast=1.0, brightness=0.5,
         wedge=0, overplot=0, aspect=0,
         doContour=0, levels=[], labelContour=0, noerase=0):
    """
    DES: do a image drawing
    INP: (map_array)  map to display
    OPT: (2elts arrays)   sizeX/Y = the 'physical' size of the array (default pixel numbers)
                                    defined by the center of the two extreme pixels !
         (2elts arrays) limitsX/Y = limits to use in X/Y for the plot
         (logical)           nan  = set if NaN are present in the array
         (string)         labelX  = x label (default 'x')
         (string)         labelY  = y label (default 'y')
         (string)         caption = the caption of the plot (default ' ')
         (char)           style   = the color used for the plot (default 'g2r'
                                    see BogliPlot.Plot.setImaCol())
         (logical)        wedge   = shall we draw a wedge ? (default no)
         (logical)         aspect = keep the aspect ratio in 'physical' unit
         (logical)       overplot = are we overplotting ? (default no)
         (logical)     doContour  = draw contour instead of map (default no)
         (array)         levels   = the levels for the contours (default nContour
                                    withing plotLimitsZ)
         (logical)   labelContour = label the contours (default no)
         (logical)        noerase = do not clear the window
    """

    myTiming = Timing()

    global wasWedge, wasAspect, transformationMatrix, zAxis, contour, numPlot

    # Check for an open device first
    if not DeviceHandler.CurrentDev:
        __MessHand.warning('no device open')
        DeviceHandler.openDev()
    else:
        # to overcome the closing-without-closing feature of apexCalibrator
        DeviceHandler.__checkReopen()

    lmap_array = np.array(copy.copy(map_array))

    # points in the map that were not observed are at NaN,
    # retrieved them... ugly
    if nan:
        lmap_array = replacenan(lmap_array)

    # If the size of the map are not gived use a integer range starting
    # at 0

    if sizeX == []:
        sizeX = np.array([0.5, lmap_array.shape[0] - 0.5], np.float)
    else:
        # Just to be sure that sizeX is an array of np.float
        sizeX = np.array(sizeX, np.float)

    # Same for sizeY
    if sizeY == []:
        sizeY = np.array([0.5, lmap_array.shape[1] - 0.5], np.float)
    else:
        sizeY = np.array(sizeY, np.float)

    # Compute the transformation matrix for this image even in case of overplot

    setMapTransformation(lmap_array, sizeX=sizeX, sizeY=sizeY, WCS=WCS)

    if not overplot:
        if not noerase:
            clear()

        setMapLimits(
            lmap_array,
            limitsX=limitsX,
            limitsY=limitsY,
            limitsZ=limitsZ)
        setLabels(labelX=labelX, labelY=labelY, caption=caption)
        setImaCol(
            style=style,
            contrast=contrast,
            brightness=brightness,
            nan=nan)

        # Keep in mind if there was a wedge
        wasWedge = wedge
        wasAspect = aspect
        numPlot = 1

        setViewPoint(aspect=aspect, wedge=wedge)

    if not doContour:
        pgimag(
            np.transpose(lmap_array), lmap_array.shape[0], lmap_array.shape[1],
            0, lmap_array.shape[0] - 1, 0, lmap_array.shape[1] - 1,
            zAxis['limits'][0], zAxis['limits'][1],
            transformationMatrix)
    else:
        # set the line style first
        pgsch(contour['charheight'])  # set character height
        pgsci(contour['color'])       # set colour index
        pgslw(contour['linewidth'])   # set line width

        # If no levels have been set compute them
        if levels == []:
            nContour = 10
            levels = np.arange(nContour * 1.0) / (nContour + 1) *\
                (zAxis['limits'][1] - zAxis['limits'][0]) + zAxis[
                    'limits'][0]

        # Draw the contour
        pgcont(
            np.transpose(lmap_array), lmap_array.shape[0], lmap_array.shape[1],
            0, lmap_array.shape[0] - 1, 0, lmap_array.shape[1] - 1,
            np.array(levels), len(levels),
            transformationMatrix)

        if labelContour == 1:
            # Label the contour
            for level in np.array(levels):
                pgconl(
                    np.transpose(lmap_array), lmap_array.shape[
                        0], lmap_array.shape[1],
                    0, lmap_array.shape[0] - 1, 0, lmap_array.shape[1] - 1,
                    level, transformationMatrix, str("%7.3f" % level), 20, 10)

    if not overplot:
        plotBox()
        labelling(wedge=wedge)

    __MessHand.debug("draw done in " + str(myTiming))

# -------------------------------------------------------------------------


def panels(nX, nY):
    """
    DES: divide the plot window in nX x nY sub-panels
    INP: (int) nX,nY = number of panels in X and Y
    """
    pgsubp(nX, nY)

# -------------------------------------------------------------------------


def nextpage():
    """
    DES: start a new page or go to next panel
    """
    pgpage()

# -------------------------------------------------------------------------


def getpix(x, y):
    """
    DES: get pixel value using the mouse pointer
    """
    x, y, char = pgband(7, 0, x, y)
    return x, y, char
# --------------------------------------------------------------------------------
# Polygon methods
# -------------------------------------------------------------------------


def defPolygon():
    """
    DES: define a polygon
    """

    # Check for an open device first
    if not DeviceHandler.CurrentDev:
        __MessHand.warning('no device open')
        DeviceHandler.openDev()

    poly = []
    cursPos = pgband(7, 0, 0, 0)
    startPos = cursPos
    poly.append((cursPos[0], cursPos[1]))
    __MessHand.info(' x = ' + repr(cursPos[0]) + ' y = ' + repr(cursPos[1]))
    if cursPos[2] == 'X':
        __MessHand.warning('not enough vertices')
    else:
        pgmove(startPos[0], startPos[1])
        cursPos = pgband(1, 0, cursPos[0], cursPos[1])
        poly.append((cursPos[0], cursPos[1]))
        __MessHand.info(
            ' x = ' + repr(cursPos[0]) + ' y = ' + repr(cursPos[1]))
        if cursPos[2] == 'X':
            __MessHand.warning('not enough vertices')
        else:
            pgdraw(cursPos[0], cursPos[1])
            while True:
                cursPos = pgband(1, 0, cursPos[0], cursPos[1])
                if cursPos[2] == 'X':
                    pgdraw(startPos[0], startPos[1])
                    break
                poly.append((cursPos[0], cursPos[1]))
                __MessHand.info(
                    ' x = ' + repr(cursPos[0]) + ' y = ' + repr(cursPos[1]))
                pgdraw(cursPos[0], cursPos[1])
                pgmove(cursPos[0], cursPos[1])
    poly = np.array(poly)
    return poly
# -------------------------------------------------------------------------


def savePolygon(poly=np.zeros((1, 2)), outFile='polygon1.pol'):
    """
    DES: save polygon to a file
    """
    try:
        f = file(outFile, 'w')
    except IOError:
        __MessHand.error('could not open file %s in write mode' % (outFile))
        return
    try:
        cPickle.dump(poly, f)
        f.close()
    except:
        __MessHand.error('cannot save polygon')
    return

# -------------------------------------------------------------------------


def loadPolygon(inFile='polygon1.pol'):
    """
    DES: load polygon from a file
    """
    try:
        f = file(inFile, 'r')
    except IOError:
        __MessHand.error('could not open file %s in read mode' % (inFile))
        return
    try:
        poly = cPickle.load(f)
        f.close()
    except:
        __MessHand.error('cannot load polygon')
        return
    return poly
# -------------------------------------------------------------------------


def plotPolygon(poly=np.zeros((1, 2))):
    """
    DES: plot a polygon
    """
    pgline(poly[:, 0], poly[:, 1])
    pgdraw(poly[0, 0], poly[0, 1])

# @}



import random
import os

try:
    laboca = os.getenv('REA_HOME_LABOCA')
    exec(
        compile(open(laboca + '/boaLaboca.py').read(),
                laboca + '/boaLaboca.py',
                'exec'))
except IOError:
    print "W: $REA_HOME_LABOCA/boaLaboca.py not found"
    print "W: This means, some processing steps will fail"

#
# Functions to determine ON and OFF phases
#


def flagPhaseOn():
    chref = data.ReceiverArray.RefChannel
    flagPosition(channel=[chref], radius=20, flag=8)
    mask = data.FlagHandler.isSetMask([8], dim=1, index=chref - 1)
    chanList = data.ReceiverArray.checkChanList([])
    chanListIndexes = data.ReceiverArray.getChanIndex(chanList)
    for chan in chanListIndexes:
        data.FlagHandler.setOnMask(mask, 8, dim=1, index=chan)


def flagPhaseOff():
    chref = data.ReceiverArray.RefChannel
    flagPosition(channel=[chref], radius=20, flag=8, outer=1)
    mask = data.FlagHandler.isSetMask([8], dim=1, index=chref - 1)
    chanList = data.ReceiverArray.checkChanList([])
    chanListIndexes = data.ReceiverArray.getChanIndex(chanList)
    for chan in chanListIndexes:
        data.FlagHandler.setOnMask(mask, 8, dim=1, index=chan)


def flagOnePhase(side=0):
    # flags all receivers when the wobbler is on one side
    # INP: (int) side, 0 for left, 1 for right
    #
    # this has to be done per subscan
    nbSub = data.ScanParam.NObs
    mask = array([], 'i')
    for i in range(nbSub):
        subNum = data.ScanParam.SubscanNum[i]
        # get all Az offsets for that subscan, and compute mean
        azo = data.ScanParam.get('azoff', subscans=[subNum])
        meanAz = fStat.f_mean(azo)
        # now determine what is left, what is right
        # But we need to work on all datapoints, including flagged ones
        azo = data.ScanParam.get('azoff', subscans=[subNum], flag='None')
        maskR = np.greater(azo, meanAz)
        maskL = np.less(azo, meanAz)
        if side:
            mask = np.concatenate((mask, maskR))
        else:
            mask = np.concatenate((mask, maskL))

    chanList = data.ReceiverArray.checkChanList([])
    chanListIndexes = data.ReceiverArray.getChanIndex(chanList)
    for chan in chanListIndexes:
        data.FlagHandler.setOnMask(mask, 8, dim=1, index=chan)


def flagPhaseLeft():
    # flags all receivers when the wobbler is on left position
    flagOnePhase(side=0)


def flagPhaseRight():
    # flags all receivers when the wobbler is on right position
    flagOnePhase(side=1)


#-------------------------------------------------------------
# Here are the functions called when processing one scan
#
def calibrate(data, tau):
    mjdref = data.ScanParam.MJD[0]
    CntstoJy(data)
    rcp = getLabocaRCP(mjdref)
    data.ReceiverArray.updateRCP(rcp)
    # correctBlind(data)
    flagRCP(rcp)
    cross = getLabocaCross(mjdref)
    data.flagChannels(cross)
    data.flatfield()
    data.correctOpacity(tau)
    scanel = fStat.f_mean(data.ScanParam.get('El'))
    taucorr = exp(tau / sin(scanel * pi / 180.))
    return taucorr


def flagCable(data, cable='M'):
    # flag channels corresponding to one cable
    filename = os.path.join(os.getenv('BOA_HOME_RCP'),
                            'channels_full_set.20070509.txt')
    data.ReceiverArray.readAdditionnalIndexFile(filename, 0, 5, '#')
    chanList = data.ReceiverArray.selectAdditionnalIndex(cable)
    data.flagChannels(chanList)


def process(data, weak=0):
    # Standard reduction of data, before computing on-off
    # flag cables A and M
    flagCable(data, 'M')
    flagCable(data, 'A')

    # flag all data taken while wobbler is moving
    refch = data.ReceiverArray.RefChannel
    data.flagSpeed(above=30.)
    # zero-order baseline per subscan
    data.polynomialBaseline(order=0, subscan=1)

    if weak == 1:
        data.flagFractionRms(ratio=5)
        data.despike()
        data.medianNoiseRemoval(computeFF=1, chanRef=-1, nbloop=1, factor=0.2)
        data.despike()
        data.flagFractionRms(ratio=3)
        data.medianNoiseRemoval(computeFF=1, chanRef=-1, nbloop=3, factor=0.7)
        correlgroup(data, nbloop=5, factor=0.9)
        data.flagFractionRms(ratio=3)
        data.despike()
        data.polynomialBaseline(order=1, subscan=1)
        # lowfreq = 1./data.ScanParam.WobCycle-0.2*1./data.ScanParam.WobCycle
        # upfreq  = 1./data.ScanParam.WobCycle+0.2*1./data.ScanParam.WobCycle
        # data.blankFreq(channel=refch,below=lowfreq)
        # data.blankFreq(channel=refch,above=upfreq)
        # data.blankFreq(below=0.75/data.ScanParam.WobCycle)
        data.flattenFreq(below=0.5 / data.ScanParam.WobCycle,
                         hiref=0.7 / data.ScanParam.WobCycle)
        data.polynomialBaseline(order=0, subscan=1)
        data.despike()
    else:
        data.flagPosition(radius=20)
        data.flagFractionRms(ratio=5)
        data.despike()
        data.medianNoiseRemoval(computeFF=1, chanRef=-1, nbloop=1, factor=0.2)
        data.despike()
        data.flagFractionRms(ratio=3)
        data.medianNoiseRemoval(computeFF=1, chanRef=-1, nbloop=3, factor=0.7)
        data.despike()
        data.flagFractionRms(ratio=3)
        correlgroup(data, nbloop=2, factor=0.8)
        data.polynomialBaseline(order=1, subscan=1)
        data.unflag(flag=8)

    data.unflagChannels([refch])

#-------------------------------------------------------------


def reduceOO(ScanNr=0, tau=0.3, subscans=[], weak=1):
    """
    NAM: reduceOO (function)
    DES: this function processes one single ON-OFF scan
    INP: (int) ScanNr      : scan number
         (float) tau       : zenith opacity
         (i list) subscans : list of subscans to consider (optional,
                             default = all subscans)
         (bool) weak       : processing optimised for weak sources? (def.=1)
    """

    tst = read(ScanNr, subscans=subscans)
    if tst:
        return -999, -999, -999, -999, -999

    # define result arrays: phase_dif + dispersion for each subscan
    result, sigma, inttime = [], [], []
    count = 0
    refch = data.ReceiverArray.RefChannel
    rmson = []
    rmsoff = []
    rmsdiff = []

    # calibrate the data
    taucorr = calibrate(data, tau)
    scantime = (data.ScanParam.MJD[-1] - data.ScanParam.MJD[0]) * 24 * 3600
    process(data, weak=weak)
    data.signal(refch)

    # Store list of "off" bolo
    offChanList = data.ReceiverArray.checkChanList([])  # all valid channels
    offChanList = offChanList.tolist()
    offChanList.remove(refch)
    nbChan = len(offChanList) + 1  # including the "on" bolo

    # build arrays with flux, time and flags for the full scan
    flux = data.getChanData('flux', refch, flag='None')
    offFluxes = data.getChanListData('flux', offChanList, dataFlag='None')
    allFluxes = [flux]
    allFluxes.extend(offFluxes)
    allFluxes = array(allFluxes)
    tt = data.getChanData('mjd', refch, flag='None')

    # compute arrays of flags corresponding to each phase
    flagPhaseOn()
    flagOff = copy.copy(data.getChanData('flag', refch, flag='None'))
    data.signal(refch, overplot=1, ci=2, style='p')
    data.unflag(flag=8)

    flagPhaseOff()
    flagOn = copy.copy(data.getChanData('flag', refch, flag='None'))
    data.signal(refch, overplot=1, ci=3, style='p')
    data.unflag(flag=8)

    # now starts loop on subscans
    nbSub = len(data.ScanParam.SubscanNum)
    subInd = data.ScanParam.SubscanIndex
    subNum = data.ScanParam.SubscanNum
    for sub in range(nbSub):
        # is it nodding A or B?
        subtype = data.ScanParam.WobblerSta[subInd[0, sub]]
        #            print "Subscan: %i  - type: %s"%(subNum[sub],str(subtype))

        if subNum[sub] % 2 == 1:   # if subscans is odd, reset list of phase_diff and errors:
            tmpRes, dtmpRes = [[]], [
                []]  # to store phase_dif and error for this subscan
            for c in range(nbChan - 1):
                tmpRes.append([])
                dtmpRes.append([])
            tmptime = []
            count = count + 1

        # here starts the loop that computes each on-off
        nMax = subInd[1, sub]
        index = subInd[0, sub]

        while(index < nMax - 1):
            timeOn = []
            timeOff = []
            while(flagOn[index] and index < nMax - 1):
                index += 1
            # here is the 1st value for "on"
            onStart = index
            while(index < nMax and not flagOn[index]):
                timeOn.append(tt[index])
                index += 1
            # that was the end of this "on" serie
            onEnd = index

            if index < nMax:
                # not yet at the end, go to beginning of next "off"
                while(flagOff[index] and index < nMax - 1):
                    index += 1
                # here is the 1st value for "off"
                offStart = index
                while(index < nMax and not flagOff[index]):
                    timeOff.append(tt[index])
                    index += 1
                # that was the end of this "off" serie
                offEnd = index

            # if at least two values in each phase, compute mean values
            if len(timeOn) > 1 and len(timeOff) > 1:
                valOn = allFluxes[:, onStart:onEnd]
                valOff = allFluxes[:, offStart:offEnd]
                intOn = timeOn[-1] - timeOn[0]
                tOn = fStat.f_mean(timeOn)
                tOff = fStat.f_mean(timeOff)
                for num in range(nbChan):
                    onChan = fStat.f_median(valOn[num])
                    offChan = fStat.f_median(valOff[num])
                    rmsOn = fStat.f_rms(valOn[num], onChan)
                    rmsOff = fStat.f_rms(valOff[num], offChan)
                    # compute and store phase diff + error for all channels
                    theDiff = onChan - offChan
                    dtheDiff = sqrt(rmsOn ** 2 + rmsOff ** 2)
                    tmpRes[num].append(theDiff)
                    dtmpRes[num].append(dtheDiff / sqrt(len(valOn[num])))
                    # and only for ref. chan also stores rms on each phase
                    if num == 0:
                        sigOn = onChan
                        sigOff = offChan
                        rmson.append(rmsOn)
                        rmsoff.append(rmsOff)
                        rmsdiff.append(dtheDiff)

                tmptime.append(intOn)

                # plot all results
                plot([tOn], [sigOn], overplot=1, ci=2)
                plot([tOn, tOn], [sigOn - rmson[-1], sigOn + rmson[-1]],
                     overplot=1, ci=2, style='l')
                plot([tOff], [sigOff], overplot=1, ci=3)
                plot([tOff, tOff], [sigOff - rmsoff[-1], sigOff + rmsoff[-1]],
                     overplot=1, ci=3, style='l')

        if subNum[sub] % 2 == 0:
            meanRes = []
            sigRes = []
            sumtime = sum(tmptime)
            for num in range(nbChan):
                w = 1. / array(dtmpRes[num]) ** 2
                wsum = sum(w)
                meanRes.append(sum(array(tmpRes[num]) * w) / wsum)
                sigRes.append(sqrt(1. / wsum))

            result.append(meanRes)
            sigma.append(sigRes)
            inttime.append(sumtime)
            print "Nod-cycle: %i :" % (count)
            print " ... %6.3f +/- %5.3f Jy/b" % (meanRes[0], sigRes[0])

    ontime = sum(inttime)
    scaneff = ontime / scantime
    print "Scan time efficiency: %6.1f %% " % (100. * scaneff)
    source = data.ScanParam.Object
    mrmson = fStat.f_mean(rmson)
    mrmsoff = fStat.f_mean(rmsoff)
    mrmsdiff = fStat.f_mean(rmsdiff)

    # compute NEFD for each phase, on the full scan
    adt = (data.ScanParam.get('deltat'))
    dt = fStat.f_median(adt)
    sampfreq = 1 / dt
    print "Sample freq: %6.2f Hz" % (sampfreq)
    sendiff = mrmsdiff / taucorr * sqrt(dt) * 1000.
    senon = mrmson / taucorr * sqrt(dt) * 1000.
    senoff = mrmsoff / taucorr * sqrt(dt) * 1000.
    print "Scan NEFD: On: %6.1f, Off: %6.1f mJy sqrt(s), " % (senon, senoff)
    print "          phase difference: %6.1f mJy sqrt(s)" % (sendiff)

    return result, sigma, inttime, taucorr, source

#-------------------------------------------------------------


def doOO(scanlist=[], tau=0.3, subscans=[], weak=1,
         filename='', update=0, source='', clip=2.0):
    """
    NAM: doOO (function)
    DES: this function processes a list of ON-OFF scans
    INP: (i list) scanlist : list of scan numbers
         (float) tau       : zenith opacity (same is applied to all scans)
         (i list) subscans : optional list of subscans to consider
                            (default = all subscans)
         (bool) weak       : processing optimised for weak sources? (def.=1)
         (str) filename    : optional name of a text file, where results are
                             written out, AND from which previous results are
                             read in if the argument 'update' is set to 1
         (bool) update     : if set, read previous results from file 'filename'
         (str) source      : name of the target (used in plot titles)
         (float) clip      : when reading previous results (i.e. when update=1),
                             measurements at more than clip*sigma from the mean
                             value are discarded
    """

    # to do: use a list of taus as input

    # Initialise lists and dictionary to store results
    fluxlist = []
    dfluxlist = []
    nefdlist = []
    timelist = []
    rmsOff, meanOff = [], []
    if update:
        if os.path.exists(filename):
            fin = file(filename)
            rl = fin.readlines()
            fin.close()
            for l in rl:
                tmp = l.split()
                fluxlist.append(float(tmp[0]))
                dfluxlist.append(float(tmp[1]))
                nefdlist.append(float(tmp[2]))
                timelist.append(float(tmp[3]))
                meanOff.append(float(tmp[4]))
                rmsOff.append(float(tmp[5]))
        tmpmean = fStat.f_mean(fluxlist)
        tmprms = fStat.f_rms(fluxlist, tmpmean)
        uplim = tmpmean + clip * tmprms
        lowlim = tmpmean - clip * tmprms
        # print uplim,lowlim
        fluxlist = []
        dfluxlist = []
        nefdlist = []
        timelist = []
        rmsOff, meanOff = [], []
        if os.path.exists(filename):
            fin = file(filename)
            rl = fin.readlines()
            fin.close()
            for l in rl:
                tmp = l.split()
                test = float(tmp[0])
                if (test > lowlim) & (test < uplim):
                    # print uplim,lowlim,test
                    fluxlist.append(float(tmp[0]))
                    dfluxlist.append(float(tmp[1]))
                    nefdlist.append(float(tmp[2]))
                    timelist.append(float(tmp[3]))
                    meanOff.append(float(tmp[4]))
                    rmsOff.append(float(tmp[5]))

    fluxOff = {}
    dfluxOff = {}
    cumOff = {}
    dcumOff = {}
    src = ''

    for snum in scanlist:
        flux, dflux, tOn, taucorr, src = reduceOO(
            snum, tau, subscans, weak=weak)
        if flux != -999:
            nb = len(flux)
            nbChan = len(flux[0])
            timelist.extend(tOn)
            for i in range(nb):
                fluxlist.append(flux[i][0])
                dfluxlist.append(dflux[i][0])
                nefdlist.append(1000. * dflux[0][i] / taucorr * sqrt(tOn[i]))
                # compute weighted mean and rms of OFF signals
                offSig = flux[i][1:]
                nbOff = float(len(offSig))
                offWei = 1. / array(dflux[i][1:]) ** 2
                w_offSig = offWei * array(offSig)
                meanOff.append(sum(w_offSig) / sum(offWei))
                rmsOff.append(sum(1. / array(dflux[i][1:])) / sum(offWei))

    # Export results, if required
    entries = len(fluxlist)
    if filename and scanlist:
        # don't write file if no input scan
        fout = file(filename, 'w')
        for i in range(entries):
            oneLine = str("%10.6f %10.6f " % (fluxlist[i], dfluxlist[i]))
            oneLine += str("%6.2f %6.2f " % (nefdlist[i], timelist[i]))
            oneLine += str("%10.6f %10.6f\n" % (meanOff[i], rmsOff[i]))
            fout.write(oneLine)
        fout.close()

    # Compute NEFD and statistical errors
    meannefd = fStat.f_mean(nefdlist)
    mean = fStat.f_mean(fluxlist)
    sysrms = fStat.f_rms(fluxlist, mean)
    syserror = sysrms / sqrt(len(fluxlist))

    # Compute cumulated fluxes (ON and OFF), and their errors
    cumflux = zeros(entries, 'f')
    dcumflux = zeros(entries, 'f')
    cumoff = zeros(entries, 'f')
    dcumoff = zeros(entries, 'f')
    cumtime = zeros(entries, 'f')
    inlist = range(entries)

    # Shuffle all measurements many times...
    nbIter = 100
    for k in range(nbIter):
        outlist = reorder(inlist)
        sumtime = 0.0
        sumflux = 0.0
        sumoff = 0.0
        wonsum = 0.0
        woffsum = 0.0
        for i in range(entries):
            num = outlist[i]
            won = 1. / dfluxlist[num] ** 2
            wonsum = wonsum + won
            woff = 1. / rmsOff[num] ** 2
            woffsum = woffsum + woff
            sumtime = sumtime + timelist[num]
            sumflux = sumflux + fluxlist[num] * won
            sumoff = sumoff + meanOff[num] * woff
            cumflux[i] = cumflux[i] + sumflux / wonsum
            dcumflux[i] = dcumflux[i] + sqrt(1. / wonsum)
            cumoff[i] = cumoff[i] + sumoff / woffsum
            dcumoff[i] = dcumoff[i] + sqrt(1. / woffsum)
            cumtime[i] = cumtime[i] + sumtime

    cumflux = cumflux / float(nbIter)
    dcumflux = dcumflux / float(nbIter)
    cumoff = cumoff / float(nbIter)
    dcumoff = dcumoff / float(nbIter)
    cumtime = cumtime / float(nbIter)

    # Finally, plot the results
    toplot = cumflux.tolist()
    toplot.extend(cumflux + dcumflux)
    toplot.extend(cumflux - dcumflux)
    toplot.extend(cumoff + dcumoff)
    toplot.extend(cumoff - dcumoff)
    mini, maxi = fStat.minmax(toplot)
    xmin, xmax = fStat.minmax(cumtime)

    Plot.panels(1, 2)
    oldCharHeight = copy.copy(BogliConfig.box['charheight'])
    BogliConfig.box['charheight'] = 1.5
    BogliConfig.cLabel['charheight'] = 1.5
    BogliConfig.xAxis['charheight'] = 1.5
    BogliConfig.yAxis['charheight'] = 1.5
    BogliConfig.globalViewPoint['marging_top'] = 0.9
    BogliConfig.globalViewPoint['marging_bottom'] = 0.15

    # First plot: down-integrating results
    Plot.nextpage()
    if not source:
        source = src
    if maxi > 1.0:
        plot(cumtime, cumflux, limitsX=[xmin, xmax], limitsY=[mini, maxi],
             style='l', ci=2, labelX='ON-Time [s]', labelY='Flux [Jy]',
             caption='Cumulative on-off flux on ' + source, noerase=1)
        for i in range(len(cumflux)):
            up = (cumflux[i] + dcumflux[i])
            low = (cumflux[i] - dcumflux[i])
            plot([cumtime[i], cumtime[i]],
                 [low, up], overplot=1, ci=2, style='l')
        plot(cumtime, cumoff + dcumoff, style='l', ci=3, overplot=1)
        plot(cumtime, cumoff - dcumoff, style='l', ci=3, overplot=1)
        plot(cumtime, cumoff, style='l', ci=1, overplot=1)
    else:
        plot(cumtime, 1000.0 * array(cumflux), limitsX=[xmin, xmax],
             limitsY=[1000.0 * mini, 1000.0 * maxi], style='l', ci=2,
             labelX='ON-Time [s]', labelY='Flux [mJy]',
             caption='Cumulative on-off flux on ' + source, noerase=1)
        for i in range(len(cumflux)):
            up = 1000.0 * (cumflux[i] + dcumflux[i])
            low = 1000.0 * (cumflux[i] - dcumflux[i])
            plot([cumtime[i], cumtime[i]],
                 [low, up], overplot=1, ci=2, style='l')
        plot(
            cumtime,
            1000. *
            cumoff +
            1000. *
            dcumoff,
            style='l',
            ci=3,
            overplot=1)
        plot(
            cumtime,
            1000. *
            cumoff -
            1000. *
            dcumoff,
            style='l',
            ci=3,
            overplot=1)
        plot(cumtime, 1000. * cumoff, style='l', ci=1, overplot=1)

    # and print the final numbers
    print "###################################################"
    if cumflux[-1] > 1.:
        print "Final result = %6.3f +/- %6.3f [+/- %6.3f] Jy/b" % (cumflux[-1], dcumflux[-1], syserror)
    else:
        print "Final result = %6.2f +/- %6.2f [+/- %6.2f] mJy/b" % (1000. * cumflux[-1], 1000. * dcumflux[-1], 1000. * syserror)

    print "Average OFF bolo = %6.2f +/- %5.1f mJy/b" % (1000. * cumoff[-1], 1000. * dcumoff[-1])
    print "Effective NEFD: %6.1f mJy sqrt(s)" % (meannefd)
    print "###################################################"

    # Second plot: show individual data points with error bars
    Plot.nextpage()
    valmin = array(fluxlist) - array(dfluxlist)
    valmax = array(fluxlist) + array(dfluxlist)
    sumtime = array(timelist)
    for i in range(entries):
        sumtime[i] = sum(timelist[:i + 1])
    ymin, ymax = fStat.minmax(np.concatenate([valmin, valmax]))
    pointSize(6)
    if ymax > 1.0:
        plot(sumtime, fluxlist, noerase=1, labelX='ON-Time [s]',
             labelY='Flux [Jy]', limitsY=[ymin, ymax],
             caption='Individual ON-OFF measurements', ci=2)
        for i in range(entries):
            plot([sumtime[i], sumtime[i]], [valmin[i], valmax[i]],
                 overplot=1, style='l', ci=2)
    else:
        plot(sumtime, 1000. * array(fluxlist), noerase=1, labelX='ON-Time [s]',
             labelY='Flux [mJy]', limitsY=[1000. * ymin, 1000. * ymax],
             caption='Individual ON-OFF measurements', ci=2)
        valmin *= 1000.
        valmax *= 1000.
        for i in range(entries):
            plot([sumtime[i], sumtime[i]], [valmin[i], valmax[i]],
                 overplot=1, style='l', ci=2)

    # Reset some graphic parameters to previous values
    Plot.panels(1, 1)
    BogliConfig.box['charheight'] = oldCharHeight
    BogliConfig.cLabel['charheight'] = oldCharHeight
    BogliConfig.xAxis['charheight'] = oldCharHeight
    BogliConfig.yAxis['charheight'] = oldCharHeight

#----------------------------------------------------------------------
# Some utilities
#


def reorder(inlist):
    nb = len(inlist)
    for i in range(nb):
        j = random.randrange(i, nb)
        inlist[i], inlist[j] = inlist[j], inlist[i]

    for i in range(nb):
        j = random.randrange(i, nb)
        inlist[i], inlist[j] = inlist[j], inlist[i]

    for i in range(nb):
        j = random.randrange(i, nb)
        inlist[i], inlist[j] = inlist[j], inlist[i]

    return inlist


def flagElJitter(scan, deltaEl=data.ReceiverArray.BeamSize / 5.0):
    read(scan)
    data.flagSpeed(above=30.)
    flagPosition(radius=20, outer=1, flag=8)
    eloff = data.getChanData('eloff')
    ondumps = len(eloff)
    melof = fStat.f_mean(eloff)
    elrms = fStat.f_rms(eloff, melof)
    low = melof - deltaEl
    up = melof + deltaEl
    data.flag(
        dataType='eloff',
        channel=data.ReceiverArray.RefChannel,
     below=low,
     flag=3)
    data.flag(
        dataType='eloff',
        channel=data.ReceiverArray.RefChannel,
     above=up,
     flag=3)
    remaindumps = len(eloff)
    elrms = fStat.f_rms(eloff, melof)
    fracflag = 100.0 * (1.0 - 1.0 * remaindumps / ondumps)
    unflag(flag=8)
    print "%4.1f %% of on dumps flagged" % (fracflag)
    print "remaining rms in elevation tracing: %4.1f arcsec" % (elrms)



# Copyright (C) 2002-2010
# Max-Planck-Institut fuer Radioastronomie Bonn
# Argelander Institut fuer Astronomie
# Astronomisches Institut der Ruhr-Universitaet Bochum
#
# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#

"""
NAM: BogliMultiChanPlot.py (module)
DES: contains Bogli multi-channel plot class
"""
__version__ = '$Revision: 2722 $'
__date__ = '$Date: 2010-04-29 11:43:32 +0200 (jeu. 29 avril 2010) $'

# ------------------------------------------------------------------------------
# ----- Import -----------------------------------------------------------------
# ------------------------------------------------------------------------

from ppgplot import *
import numpy as np

from . import Plot
from rea import ReaMessageHandler
from rea.Utilities import Timing
from rea.fortran.fUtilities import mreplacenan, masknan

from rea.fortran.fStat import minmax

from . import DeviceHandler

# ------------------------------------------------------------------------------
# ----- MultiChan --------------------------------------------------------------
# ------------------------------------------------------------------------

# Description of the Channel Number, same definition as Plot.__labelX
chanNum = {'charheight': 1.0,
           'color': 2,
           'displacement': -1.5,
           'coordinate': 0.1,
           'justification': 0.0,
           'side': 't',
           'text': ' ',
           'draw_it': 1}


# The individuals viewPoint size, saved for overplot
viewPointStep = {'X': 1.0, 'Y': 1.0}

# The private message Handler
__MessHand = ReaMessageHandler.MessHand('MultiPlot')

# ----------------------------------------------------------------------------
# ----- Methods --------------------------------------------------------------
# ----------------------------------------------------------------------------


def detSubDivView(numPlot, wedge=0):
    """
    DES: Determine sub-division of plot page into viewpoints/boxes.
    INP: (int) numPlot   : number of panels in the plot
    OPT: (logical) wedge : if a wedge is present (no by default)
    OUT: (ints) (numPlotWinX,numPlotWinY) the number of viewport in both directions
    """

    global viewPointStep, chanNum

    # Divide in a square
    numPlotWinX = np.int(np.floor(np.sqrt(numPlot)))
    numPlotWinY = np.int(np.floor(numPlot / numPlotWinX))

    # In case the is more window than a perfect square
    if numPlotWinX * numPlotWinY < numPlot:
        numPlotWinY += 1

    if wedge == 1 or Plot.wasWedge == 1:
        # If wedge is present reserve space for it
        viewPointStep['X'] = (Plot.globalViewPoint['marging_right'] -
                              Plot.globalViewPoint['marging_left'] -
                              Plot.globalViewPoint['wedge_size']) / \
            numPlotWinX
    else:
        viewPointStep['X'] = (Plot.globalViewPoint['marging_right'] -
                              Plot.globalViewPoint['marging_left']) / \
            numPlotWinX

    viewPointStep['Y'] = (Plot.globalViewPoint['marging_top'] -
                          Plot.globalViewPoint['marging_bottom']) / \
        numPlotWinY

    # Reserve room for the wedge if needed
    if wedge:
        viewPointStep['X'] -= Plot.globalViewPoint['wedge_size'] / numPlotWinX

    # Change the character height of the plot according to the
    # computed viewPoint size
#  box['charheight']     = 1.0/np.sqrt(0.7*numPlotWinY)
    chanNum['charheight'] = 1.0 / np.sqrt(0.3 * numPlotWinY)

    # TODO: somehow the Point['size'] should be change
    # Point['size'] = 4.0/np.sqrt(0.7*numPlotWinY)

    return (numPlotWinX, numPlotWinY)

# ----------------------------------------------------------------------------


def plotBox(numPlot, x, y):
    """
    DES: Draw box and labels.
    INP: (int) numPlot : number of panels in the plot
         (int) x,y     : the indices of the panel
    """

    myTiming = Timing()

    pgsch(Plot.box['charheight'] / np.sqrt(0.5 * np.sqrt(numPlot)))
          # set character height
    pgsci(Plot.box['color'])                                 # set colour index
    pgsls(Plot.box['linestyle'])                             # set line style
    pgslw(Plot.box['linewidth'])                             # set line width
    lastRow = 0

    if (1 + y) * np.floor(np.sqrt(numPlot)) >= numPlot:
        lastRow = 1

    if Plot.xAxis['log']:
        codeX = 'BCTSL'
    else:
        codeX = 'BCTS'
    if Plot.yAxis['log']:
        codeY = 'BCTSL'
    else:
        codeY = 'BCTS'

    if x == 0 and lastRow:
        pgbox(codeX + 'N', 0.0, 0, codeY + 'NV', 0.0, 0)
              # corner: label x and y
    elif x == 0:
        pgbox(codeX, 0.0, 0, codeY + 'NV', 0.0, 0)                  # label y
    elif lastRow:
        pgbox(codeX + 'N', 0.0, 0, codeY, 0.0, 0)                   # label x
    else:
        pgbox(codeX, 0.0, 0, codeY, 0.0, 0)
              # label nothing

    __MessHand.debug("plotBox done in " + str(myTiming))

# -------------------------------------------------------------------------


def setLimits(dataX, dataY, limitsX=[], limitsY=[]):
    """
    DES: compute and/or set the limits for the multiplot
    INP: (list arrays)   dataX/Y : the array to be plotted
    OPT: (2elts array) limitsX/Y : limits to use in X/Y for the plot
    """

    myTiming = Timing()

    xLimits = []
    yLimits = []

    # compute the limits for each plot and ...
    for i in range(len(dataX)):
        if len(dataX[i]) > 0:
            # Make sure that X and Y have the same length
            if len(dataX[i]) > len(dataY[i]):
                dataX[i] = dataX[i][:len(dataY[i])]
            elif len(dataY[i]) > len(dataX[i]):
                dataY[i] = dataY[i][:len(dataX[i])]

            # Remove all NaN from X and Y arrays
            maskX = np.isnan(dataX[i])
            maskY = np.isnan(dataY[i])
            mNan = np.bitwise_or(maskX, maskY)
            bad = np.where(mNan)[0]
            if len(bad):
                # at least one NaN somewhere
                __MessHand.warning(
                    "%i data points NaN were found" %
                    (len(bad)))
                good = np.nonzero(np.equal(mNan, 0))[0]
                if len(good):
                    dataX[i] = np.take(dataX[i], good)
                    dataY[i] = np.take(dataY[i], good)
                else:
                    dataX[i] = np.array([0], 'f')
                    dataY[i] = np.array([0], 'f')

            Plot.setLimits(
                dataX[i],
                dataY[i],
                limitsX=limitsX,
                limitsY=limitsY)
            xLimits.append(Plot.xAxis['limits'])
            yLimits.append(Plot.yAxis['limits'])

    xLimits = np.ravel(xLimits)
    yLimits = np.ravel(yLimits)

    # .. np.take the maximum of all
    Plot.xAxis['limits'] = minmax(xLimits)
    Plot.yAxis['limits'] = minmax(yLimits)

    __MessHand.debug("setLimits time: " + myTiming.__str__())


# -------------------------------------------------------------------------
def drawChanNum(c):
    """
    DES: Draw channel number.
    INT: (int) c : the channel number
    """

    global chanNum

    chanNum['text'] = str(c)
    Plot.drawLabel(chanNum)

# -------------------------------------------------------------------------


def setMultiViewPoint(x=0, y=0, wedge=0):
    """
    DES: Determine and set view points.
    INP: (ints)    x/y      : position of the viewpoint from 0 to numPlotWinX/Y
    """

    global viewPointStep

    xleftIndVP = Plot.globalViewPoint[
        'marging_left'] + x * viewPointStep[
            'X']
    xrightIndVP = Plot.globalViewPoint[
        'marging_left'] + (x + 1.0) * viewPointStep['X']
    ytopIndVP = Plot.globalViewPoint[
        'marging_top'] - y * viewPointStep[
            'Y']
    ybotIndVP = Plot.globalViewPoint[
        'marging_top'] - (y + 1.0) * viewPointStep['Y']

    # Set the viewpoint...
    pgsvp(xleftIndVP, xrightIndVP, ybotIndVP, ytopIndVP)

    # ... and set the limits for the plots
    pgswin(Plot.xAxis['limits'][0], Plot.xAxis['limits'][1],
           Plot.yAxis['limits'][0], Plot.yAxis['limits'][1])

    # TODO : Do we need an option for aspect ratio in multichan plots ?
# -------------------------------------------------------------------------


def gloLabelling(wedge=0):
    """
    DES: Label x, y, caption and channel number.
    OPT: (logical) wedge : if a wedge is present
    """

    # Use the ViewPoint as if there were only one plot
    Plot.setViewPoint(wedge=wedge)
    # And label it.
    Plot.labelling(wedge=wedge)

# -------------------------------------------------------------------------


def plot(chanList, dataX, dataY,
         limitsX=[], limitsY=[],
         labelX='x', labelY='y', caption=' ',
         style='p', ci=1, overplot=0, logX=0, logY=0,
         nan=0, noerase=0):
    """
    DES: do a multi channel plot
    INP: (int list) chanList = list of channels
         (array list)  dataX = values to plot along X
         (array list)  dataY = values to plot along Y
    OPT: (2elts array) limitsX/Y = limits to use in X/Y for the plot
         (string)      labelX  = x label (default 'x')
         (string)      labelY  = y label (default 'y')
         (string)      caption = the caption of the plot (default ' ')
         (char)        style   = the style used for the plot ('l': line,
                                 'p': point (default), 'b': histogram)
         (int)            ci   = color index (default 1)
         (logical)    overplot = are we overplotting ? (default no)
         (logical)      logX/Y = do we use log scale ? (default no)
         (logical)     noerase = do not clear the window
    """
    myTiming = Timing()

    # Check for an open device first
    if not DeviceHandler.CurrentDev:
        __MessHand.warning('no device open')
        DeviceHandler.openDev()

    if overplot:
        if len(chanList) != Plot.numPlot:
            __MessHand.error('cannot overplot, wrong number of channels')
            return
    else:
        # If we are not overplotting clear the graph and set the limits/label
        if not noerase:
            Plot.clear()
        if logX:
            Plot.xAxis['log'] = 1
        if logY:
            Plot.yAxis['log'] = 1
        setLimits(dataX, dataY, limitsX=limitsX, limitsY=limitsY)
        Plot.setLabels(labelX=labelX, labelY=labelY, caption=caption)
        Plot.numPlot = len(chanList)

    # Divide the screen properly
    (numPlotWinX, numPlotWinY) = detSubDivView(Plot.numPlot)

    # pgbbuf() buffers the drawing on the screen until the pgebuf() command.
    # There is a gain of up to 10% in time when using this on a
    # multichanplot of 400 channels on full screen resolution.

    pgbbuf()

    # Now do the plots
    for y in range(numPlotWinY):
        for x in range(numPlotWinX):

            # counter of plotted channel
            cc = y * numPlotWinX + x

            # terminate when last chan to plot is reached
            if cc == Plot.numPlot:
                break

            # determine and set view point
            setMultiViewPoint(x, y)

            if not overplot and Plot.box['draw_it']:   # draw box and labels
                plotBox(Plot.numPlot, x, y)

            # Do the plot
            Plot.plotDataXY(dataX[cc], dataY[cc], style=style, ci=ci)

            if not overplot and chanNum['draw_it']:  # draw the channel number
                drawChanNum(chanList[cc])

    if not overplot:
        # do not forget the label everything if we are not overplotting
        gloLabelling()

    pgebuf()

    __MessHand.debug("plot done in " + str(myTiming))

# -------------------------------------------------------------------------


def draw(chanList, map_arrays, sizeX=[], sizeY=[], WCS=[],
         limitsX=[], limitsY=[], limitsZ=[],
         nan=0,
         labelX='x', labelY='y', caption=' ',
         style='g2r', contrast=1.0, brightness=0.5,
         wedge=0, overplot=0):
    """
    DES: do a multi channel image drawing
    INP: (int list) chanList = list of channels
         (map_arrays)  lits of map to display
    OPT: (2elts arrays)   sizeX/Y = the 'physical' size of the array (default pixel numbers)
         (2elts arrays) limitsX/Y = limits to use in X/Y for the plot
         (string)         labelX  = x label (default 'x')
         (string)         labelY  = y label (default 'y')
         (string)         caption = the caption of the plot (default ' ')
         (char)           style   = the color used for the plot (default 'g2r'
                                    see Plot.Plot.setImaCol())
         (logical)        wedge   = shall we draw a wedge ? (default no)
         (logical)       overplot = are we overplotting ? (default no)
    """

    myTiming = Timing()

    # Check for an open device first
    if not DeviceHandler.CurrentDev:
        __MessHand.warning('no device open')
        DeviceHandler.openDev()

    lmap_arrays = np.array(copy.copy(map_arrays))

    # points in the map that were not observed are at NaN,
    # retrieved them... ugly
    if nan:
        new = mreplacenan(lmap_arrays)
        for i in range(len(lmap_arrays)):
            lmap_arrays[i] = new[i]

    # If the size of the map have not been given use a integer
    # range, use the lmap_arrays[0], i.e. all the maps have to
    # have the same dimension for the moment
    # TODO: fix that

    if sizeX == []:
        sizeX = np.array([0, lmap_arrays[0].shape[0] - 1], np.float)
    else:
        # Just to be sure that sizeX is an array of np.float
        sizeX = np.array(sizeX, np.float)

    # Same for sizeY
    if sizeY == []:
        sizeY = np.array([0, lmap_arrays[0].shape[1] - 1], np.float)
    else:
        sizeY = np.array(sizeY, np.float)

    # Compute the transformation matrix for this image even in case of overplot

    Plot.setMapTransformation(lmap_arrays, sizeX=sizeX, sizeY=sizeY, WCS=WCS)

    if not overplot:
        # Since we are note overplotting, clear the plot, set
        # size/limits/labels/color

        Plot.clear()

        Plot.setMapLimits(
            lmap_arrays,
            limitsX=limitsX,
            limitsY=limitsY,
            limitsZ=limitsZ)
        Plot.setLabels(labelX, labelY, caption)
        Plot.setImaCol(style=style, contrast=contrast, brightness=brightness)
        Plot.numPlot = len(chanList)
        # Keep in min that they were a wedge
        Plot.wasWedge = wedge

    if overplot:
        if len(chanList) != numPlot:
            __MessHand.error('cannot overplot, wrong number of channels')
            return

    # Divide the screen properly
    (numPlotWinX, numPlotWinY) = detSubDivView(Plot.numPlot, wedge=wedge)

    # pgbbuf() buffers the drawing on the screen til the pgebuf()
    # command.
    pgbbuf()

    # Now do the plots
    for y in range(numPlotWinY):
        for x in range(numPlotWinX):

            # Channel counter
            cc = y * numPlotWinX + x

            # terminate when last chan to plot is reached
            if cc == Plot.numPlot:
                break

            # determine and set view point
            setMultiViewPoint(x, y)

            if Plot.box['draw_it']:   # draw box and labels
                plotBox(Plot.numPlot, x, y)

            # Draw the image
            pgimag(
                np.transpose(lmap_arrays[cc]), lmap_arrays[
                    cc].shape[0], lmap_arrays[cc].shape[1],
                0, lmap_arrays[cc].shape[
                    0] - 1, 0, lmap_arrays[cc].shape[1] - 1,
                Plot.zAxis['limits'][0], Plot.zAxis['limits'][1],
                Plot.transformationMatrix)

            if chanNum['draw_it']:  # draw the channel number
                drawChanNum(chanList[cc])  # 0=black

    if not overplot:
        # Do not forget to label everything with a wedge if needed
        gloLabelling(wedge=wedge)

    pgebuf()

    __MessHand.debug("draw done in " + str(myTiming))


# @}



# Copyright (C) 2014
# Institut d'Astrophysique Spatiale
#
# Forked from the BoA project
#
# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#

"""
..module:: ReaFlagHandler
:synopsis: Contains classes for bitwise manipulation of flags.
     The module serves as an object-oriented interface for the Fortran
     module fFlagHandler.
"""
# ---------------------------------------------------------------------
# ---- Import ---------------------------------------------------------
# ---------------------------------------------------------------------

import copy
import numpy as np
from rea.Utilities import as_column_major_storage
from .fortran import fFlagHandler


# ---------------------------------------------------------------------
# ---- Function createFlagHandler B-------------------------------------
# ---------------------------------------------------------------------

def createFlagHandler(aFlags):
    """Creates an object of the correct subtype of FlagHandler to handle
         the flag array aFlags.  Presently, 1-dim flag arrays of type
         Int32 and 2-dim flag arrays of type Int8 can be handled.

    Parameters
    ----------
    aFlags : array
        Flag array

    Returns
    -------
    ReaFlagHangler.FlagHandler
        FlagHandler object
    """
    typecode = aFlags.dtype.name
    nDim = len(aFlags.shape)

    if nDim == 1 and typecode == 'int32':
        return FlagHandler1d32b(aFlags)
    elif nDim == 2 and typecode == 'int8':
        return FlagHandler2d8b(aFlags)
    else:
        raise Exception(
            "Invalid Flag array: dim=%d, typecode=%s" %
            (nDim, typecode))


# ---------------------------------------------------------------------
# ---- Class FlagHandler ----------------------------------------------
# ---------------------------------------------------------------------

class FlagHandler:

    """..class:: FlagHandler
    :synopsis: Provides methods to manipulate and query flag arrays
    bitwise.  Here, a flag array is a np.array, each element of which
    represents n independent flags, with n the number of bits of each
    array element. The n independent flags are enumerated with the
    flag value, ranging from 1 (the rightmost bit) to n (the leftmost
    bit).

         Throughout the class, the nomenclature is as follows:

         - self._aFlags (np.array):
             The flag array to be manipulated or queried.

             Permitted shapes and datatypes of self._aFlags depend on the
             exact subtype of FlagHandler.

         - iFlags (integer, list of integers, or empty list):
             The flag values used for manipulating or queriing the flag
             array.

             Valid flag values are from 1 to n, with n the number of bits
             of each element of aFlags.

             If iFlags is the empty list, the list [1,2,...,n] is assumed.

    """

    # -----------------------------------------------------------------

    def __init__(self, aFlags):
        self._aFlags = aFlags

        self._validFlagValues = range(1, self._aFlags.itemsize * 8 + 1)

    # -----------------------------------------------------------------

    def getFlags(self):
        """Returns the flag array self._aFlags"""
        return self._aFlags

    # -----------------------------------------------------------------

    def getValidFlagValues(self):
        """Returns a list of all valid flag values for
             flag array self._aFlags

        Returns
        -------
        list of int
            All valid flag values for aFlags
        """
        return copy.copy(self._validFlagValues)

    # -----------------------------------------------------------------

    def _getBitmask(self, iFlags):
        # DES: Determine the bitmask corresponding to iFlags.
        # INP: iFlags (int list): List of flag values that specify bits.
        # OUT: np.array of the same typecode as self._aFlags,
        #      shape(1,), and value such that bits corresponding
        #      to iFlags are 1, all others 0
        typecode = self._aFlags.dtype.char
        zero = np.zeros(1, typecode)
        one = np.ones(1, typecode)

        bitmask = zero

        if iFlags == []:
            iFlags = self._validFlagValues
        if isinstance(iFlags, type(1)):
            iFlags = [iFlags]

        for iFlag in iFlags:
            if not iFlag in self._validFlagValues:
                raise "Invalid flag %d" % iFlag
            bitmask |= (one << (iFlag - 1)).astype(typecode)

        return bitmask


class FlagHandler1d32b(FlagHandler):

    """..class:: FlagHandler1d32b
    :synopsis: Bitwise manipulation of 1-dim flag arrays of type Int32
    """

    # -----------------------------------------------------------------

    def __init__(self, aFlags):
        FlagHandler.__init__(self, aFlags)

    # -----------------------------------------------------------------

    def setAll(self, iFlags=[]):
        """Sets the flag values iFlags for all elements of self._aFlags

        Parameters
        ----------
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        """
        bitmask = self._getBitmask(iFlags)
        fFlagHandler.flags1d32b.setall(self._aFlags, bitmask)

    # -----------------------------------------------------------------

    def unsetAll(self, iFlags=[]):
        """Unsets the flag values iFlags for all elements of self._aFlags

        Parameters
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        """
        bitmask = self._getBitmask(iFlags)
        fFlagHandler.flags1d32b.unsetall(self._aFlags, bitmask)

    # -----------------------------------------------------------------

    def nSet(self, iFlags=[]):
        """Returns the number of elements of self._aFlags for which
        at least one flag value specified by iFlags is set.

        Parameters
        ----------
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)

        Returns
        -------
        int
            Number of elements of self._aFlags for which at least one
        flag value specified by iFlags is set.

        """
        bitmask = self._getBitmask(iFlags)
        return fFlagHandler.flags1d32b.nset(self._aFlags, bitmask)

    # -----------------------------------------------------------------

    def nUnset(self, iFlags=[]):
        """Returns the number of elements of self._aFlags for which
        none of the flag values specified by iFlags is set.

        Parameters
        ----------
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)

        Returns
        -------
        int
            Number of elements of self._aFlags for which none of the
        flag values specified by iFlags is set.

        """
        bitmask = self._getBitmask(iFlags)
        return fFlagHandler.flags1d32b.nunset(self._aFlags, bitmask)

    # -----------------------------------------------------------------

    def setOnIndex(self, index, iFlags=[]):
        """Sets the flag values iFlags for a single element of
        flag array self._aFlags

        Parameters
        ----------
        index : int
            Index of the element of self._aFlags to be set
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        """
        bitmask = self._getBitmask(iFlags)
        fFlagHandler.flags1d32b.setonindex(self._aFlags, index + 1, bitmask)

    # -----------------------------------------------------------------

    def unsetOnIndex(self, index, iFlags=[]):
        """Unsets the flag values iFlags for a single element of
        flag array self._aFlags

        Parameters
        ----------
        index : int
            Index of the element of self._aFlags to be unset
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        """
        bitmask = self._getBitmask(iFlags)
        fFlagHandler.flags1d32b.unsetonindex(self._aFlags, index + 1, bitmask)

    # -----------------------------------------------------------------

    def isSetOnIndex(self, index, iFlags=[]):
        """Returns 1 if at least one flag value specified by iFlags
        is set for a single element of flag array self._aFlags.

        Parameters
        ----------
        index : int
            Index of the element of self._aFlags to be set
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)

        Returns
        -------
        int
            1 if at least one flag value specified by iFlags
        is set, 0 else.
        """
        bitmask = self._getBitmask(iFlags)
        return fFlagHandler.flags1d32b.issetonindex(self._aFlags,
                                                    index + 1,
                                                    bitmask)

    # -----------------------------------------------------------------

    def isUnsetOnIndex(self, index, iFlags=[]):
        """Returns 1 if none of the flag values specified by iFlags
        are set for a single element of flag array self._aFlags.

        Parameters
        ----------
        index : int
            Index of the element of self._aFlags to be set
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)

        Returns
        -------
        int
            1 if none of the flag values specified by iFlags
        is set, 0 else.
        """
        bitmask = self._getBitmask(iFlags)
        return fFlagHandler.flags1d32b.isunsetonindex(self._aFlags,
                                                      index + 1,
                                                      bitmask)

    # -----------------------------------------------------------------

    def setOnMask(self, aMask, iFlags=[]):
        """Sets the flag values iFlags for all elements of
        flag array self._aFlags specified by aMask

        Parameters
        aMask : array
            Mask specifiing the elements of self._aFlags to be
        manipulated.  The shape of aMask must be the shape of
        self._aFlags.  Only elements of self._aFlags, for which aMask
        is not 0, will be manipulated.
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)

        """
        if aMask.shape != self._aFlags.shape:
            raise "Invalid mask"

        bitmask = self._getBitmask(iFlags)
        fFlagHandler.flags1d32b.setonmask(self._aFlags, aMask, bitmask)

    # -----------------------------------------------------------------

    def unsetOnMask(self, aMask, iFlags=[]):
        """Unsets the flag values iFlags for all elements of
        flag array self._aFlags specified by aMask

        Parameters
        aMask : array
            Mask specifiing the elements of self._aFlags to be
        manipulated.  The shape of aMask must be the shape of
        self._aFlags.  Only elements of self._aFlags, for which aMask
        is not 0, will be manipulated.
        iFlags : list of int
            Flag  values (see also doc string of class FlagHandler)

        """
        if aMask.shape != self._aFlags.shape:
            raise "Invalid mask"

        bitmask = self._getBitmask(iFlags)
        fFlagHandler.flags1d32b.unsetonmask(self._aFlags, aMask, bitmask)

    # -----------------------------------------------------------------

    def isSetMask(self, iFlags=[]):
        """Returns a mask that indicates for which elements
        of self._aFlags at least one flag value specified by
        iFlags is set.

        Parameters
        ----------
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)

        Returns
        -------
        array
            Mask indicating that at least one flag value specified by
        iFlags is set for the corresponding element of self._aFlags.

        """
        bitmask = self._getBitmask(iFlags)
        return fFlagHandler.flags1d32b.issetmask(self._aFlags, bitmask)

    # -----------------------------------------------------------------

    def isUnsetMask(self, iFlags=[]):
        """Returns a mask that indicates for which elements
        of self._aFlags none of the flag values specified by
        iFlags is set.

        Parameters
        ----------
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)

        Returns
        -------
        array
            Mask indicating that none of the flag values specified by
        iFlags is set for the corresponding element of self._aFlags.

        """
        bitmask = self._getBitmask(iFlags)
        return fFlagHandler.flags1d32b.isunsetmask(self._aFlags, bitmask)


class FlagHandler2d8b(FlagHandler):

    """..class:: FlagHandler2d8b
    :synopsis: Bitwise manipulation of 2-dim flag arrays of type Int8
    """

    # -----------------------------------------------------------------

    def __init__(self, aFlags):
        FlagHandler.__init__(self, aFlags)
        self._aFlags = as_column_major_storage(self._aFlags)

    # -----------------------------------------------------------------

    def setAll(self, iFlags=[], dim=None, index=None):
        """
        NAM: FlagHandler2d8b.setAll (Method)
        DES: Sets the flag values iFlags for all elements of self._aFlags
        INP: iFlags (int list) : Flag values (see also doc string of class FlagHandler)
             dim, index (int)  : Specify a slice of self._aFlags on which this method operates.
                                 If None (default): Method operates on complete flag array.
        """
        bitmask = self._getBitmask(iFlags)

        if dim is None:
            fFlagHandler.flags2d8b.setall(self._aFlags, bitmask)
        elif dim == 0:
            fFlagHandler.flags2d8b.setall2(self._aFlags, index + 1, bitmask)
        elif dim == 1:
            fFlagHandler.flags2d8b.setall1(self._aFlags, index + 1, bitmask)
        else:
            raise "Invalid argument 'dim'"

    # -----------------------------------------------------------------

    def unsetAll(self, iFlags=[], dim=None, index=None):
        """
        NAM: FlagHandler2d8b.unsetAll (Method)
        DES: Unsets the flag values iFlags for all elements of self._aFlags
        INP: iFlags (int list) : Flag values (see also doc string of class FlagHandler)
             dim, index (int)  : Specify a slice of self._aFlags on which this method operates.
                                 If None (default): Method operates on complete flag array.
        """
        bitmask = self._getBitmask(iFlags)

        if dim is None:
            fFlagHandler.flags2d8b.unsetall(self._aFlags, bitmask)
        elif dim == 0:
            fFlagHandler.flags2d8b.unsetall2(self._aFlags, index + 1, bitmask)
        elif dim == 1:
            fFlagHandler.flags2d8b.unsetall1(self._aFlags, index + 1, bitmask)
        else:
            raise "Invalid argument 'dim'"

    # -----------------------------------------------------------------

    def nSet(self, iFlags=[], dim=None, index=None):
        """Returns the number of elements of self._aFlags for which
        at least one flag value specified by iFlags is set.

        Parameters
        ----------
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        dim, index : int
            Specify a slice of self._aFlags on which this method operates.
            If None (default): Method operates on complete flag array.

        Returns
        -------
        int
            Number of elements of self._aFlags for which at least
        one flag value specified by iFlags is set.
        """
        bitmask = self._getBitmask(iFlags)

        if dim is None:
            n = fFlagHandler.flags2d8b.nset(self._aFlags, bitmask)
        elif dim == 0:
            n = fFlagHandler.flags2d8b.nset2(self._aFlags, index + 1, bitmask)
        elif dim == 1:
            n = fFlagHandler.flags2d8b.nset1(self._aFlags, index + 1, bitmask)
        else:
            raise "Invalid argument 'dim'"

        return n

    # -----------------------------------------------------------------

    def nUnset(self, iFlags=[], dim=None, index=None):
        """Returns the number of elements of self._aFlags for which
        none of the flag values specified by iFlags is set.

        Parameters
        ----------
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        dim, index : int
            Specify a slice of self._aFlags on which this method operates.
        If None (default): Method operates on complete flag array.

        Returns
        -------
        int
            Number of elements of self._aFlags for which none
        of the flag values specified by iFlags is set.
        """
        bitmask = self._getBitmask(iFlags)

        if dim is None:
            n = fFlagHandler.flags2d8b.nunset(self._aFlags, bitmask)
        elif dim == 0:
            n = fFlagHandler.flags2d8b.nunset2(
                self._aFlags, index + 1, bitmask)
        elif dim == 1:
            n = fFlagHandler.flags2d8b.nunset1(
                self._aFlags, index + 1, bitmask)
        else:
            raise "Invalid argument 'dim'"

        return n

    # -----------------------------------------------------------------

    def setOnIndex(self, index, iFlags=[]):
        """Sets the flag values iFlags for a single element of
        flag array self._aFlags

        Parameters
        ----------
        index : int
            Index of the element of self._aFlags to be set
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        """
        bitmask = self._getBitmask(iFlags)
        fFlagHandler.flags2d8b.setonindex(self._aFlags,
                                          index[0] + 1, index[1] + 1,
                                          bitmask)

    # -----------------------------------------------------------------

    def unsetOnIndex(self, index, iFlags=[]):
        """Sets the flag values iFlags for a single element of
        flag array self._aFlags

        Parameters
        ----------
        index : int
            Index of the element of self._aFlags to be unset
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        """
        bitmask = self._getBitmask(iFlags)
        fFlagHandler.flags2d8b.unsetonindex(self._aFlags,
                                            index[0] + 1, index[1] + 1,
                                            bitmask)

    # -----------------------------------------------------------------

    def isSetOnIndex(self, index, iFlags=[]):
        """Returns 1 if at least one flag value specified by iFlags
        is set for a single element of flag array self._aFlags.

        Parameters
        ----------
        index : int
            Index of the element of self._aFlags
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)

        Returns
        -------
        int
            1 if at least one flag value specified by iFlags
        is set, 0 else.
        """
        bitmask = self._getBitmask(iFlags)
        return fFlagHandler.flags2d8b.issetonindex(self._aFlags,
                                                   index[0] + 1, index[1] + 1,
                                                   bitmask)

    # -----------------------------------------------------------------

    def isUnsetOnIndex(self, index, iFlags=[]):
        """Returns 1 if none of the flag values specified by iFlags
        are set for a single element of flag array self._aFlags.

        Parameters
        ----------
        index : int
            Index of the element of self._aFlags
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)

        Returns
        -------
        int
            1 if none of the flag values specified by iFlags
        is set, 0 else.
        """
        bitmask = self._getBitmask(iFlags)
        return fFlagHandler.flags2d8b.isunsetonindex(self._aFlags,
                                                     index[0] + 1, index[
                                                         1] + 1,
                                                     bitmask)

    # -----------------------------------------------------------------

    def setOnMask(self, aMask, iFlags=[], dim=None, index=None):
        """Sets the flag values iFlags for all elements of
        flag array self._aFlags specified by aMask

        Parameters
        ----------
        aMask : array
            Mask specifiing the elements of self._aFlags to be
        manipulated.  The shape of aMask must be the shape of
        self._aFlags.  Only elements of self._aFlags, for which aMask
        is not 0, will be manipulated.
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        dim, index : int
            Specify a slice of self._aFlags on which this method operates.
        If None (default): Method operates on complete flag array.

        """
        bitmask = self._getBitmask(iFlags)

        if dim is None:
            fFlagHandler.flags2d8b.setonmask(self._aFlags, aMask, bitmask)
        elif dim == 0:
            fFlagHandler.flags2d8b.setonmask2(
                self._aFlags,
                aMask,
                index + 1,
                bitmask)
        elif dim == 1:
            fFlagHandler.flags2d8b.setonmask1(
                self._aFlags,
                aMask,
                index + 1,
                bitmask)
        else:
            raise "Invalid argument 'dim'"

    # -----------------------------------------------------------------

    def unsetOnMask(self, aMask, iFlags=[], dim=None, index=None):
        """Sets the flag values iFlags for all elements of
        flag array self._aFlags specified by aMask

        Parameters
        -----------
        aMask : array
            Mask specifiing the elements of self._aFlags to be
        manipulated.  The shape of aMask must be the shape of
        self._aFlags.  Only elements of self._aFlags, for which aMask
        is not 0, will be manipulated.
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        dim, index : int
            Specify a slice of self._aFlags on which this method operates.
        If None (default): Method operates on complete flag array.

        """
        bitmask = self._getBitmask(iFlags)

        if dim is None:
            fFlagHandler.flags2d8b.unsetonmask(self._aFlags, aMask, bitmask)
        elif dim == 0:
            fFlagHandler.flags2d8b.unsetonmask2(
                self._aFlags,
                aMask,
                index + 1,
                bitmask)
        elif dim == 1:
            fFlagHandler.flags2d8b.unsetonmask1(
                self._aFlags,
                aMask,
                index + 1,
                bitmask)
        else:
            raise "Invalid argument 'dim'"

    # -----------------------------------------------------------------

    def isSetMask(self, iFlags=[], dim=None, index=None):
        """Returns a mask that indicates for which elements
        of self._aFlags at least one flag value specified by
        iFlags is set.

        Parameters
        ----------
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        dim, index : int
            Specify a slice of self._aFlags on which this method operates.
        If None (default): Method operates on complete flag array.

        Returns
        -------
        array
            Mask indicating that at least one flag value specified by
        iFlags is set for the corresponding element of self._aFlags.

        """
        bitmask = self._getBitmask(iFlags)

        if dim is None:
            aMask = fFlagHandler.flags2d8b.issetmask(self._aFlags, bitmask)
        elif dim == 0:
            aMask = fFlagHandler.flags2d8b.issetmask2(
                self._aFlags, index + 1, bitmask)

        elif dim == 1:
            aMask = fFlagHandler.flags2d8b.issetmask1(
                self._aFlags, index + 1, bitmask)
        else:
            raise "Invalid argument 'dim'"

        return aMask

    # -----------------------------------------------------------------

    def isUnsetMask(self, iFlags=[], dim=None, index=None):
        """Returns a mask that indicates for which elements
        of self._aFlags none of the flag values specified by
        iFlags is set.

        Parameters
        ----------
        iFlags : list of int
            Flag values (see also doc string of class FlagHandler)
        dim, index : int
            Specify a slice of self._aFlags on which this method operates.
        If None (default): Method operates on complete flag array.

        Returns
        -------
        array
            Mask indicating that none of the flag values specified by
        iFlags is set for the corresponding element of self._aFlags.

        """
        bitmask = self._getBitmask(iFlags)

        if dim is None:
            aMask = fFlagHandler.flags2d8b.isunsetmask(self._aFlags, bitmask)
        elif dim == 0:
            aMask = fFlagHandler.flags2d8b.isunsetmask2(
                self._aFlags, index + 1, bitmask)

        elif dim == 1:
            aMask = fFlagHandler.flags2d8b.isunsetmask1(
                self._aFlags, index + 1, bitmask)
        else:
            raise "Invalid argument 'dim'"

        return aMask



# Copyright (C) 2002-2006
# Max-Planck-Institut fuer Radioastronomie Bonn
# Argelander Institut fuer Astronomie
# Astronomisches Institut der Ruhr-Universitaet Bochum
#
# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#

"""
NAM: DeviceHandler.py (module)
"""
__version__ = '$Revision: 2381 $'
__date__ = '$Date: 2007-08-29 10:06:23 +0200 (mer, 29 aou 2007) $'
#----------------------------------------------------------------------------------
#----- Import ---------------------------------------------------------------------
#-------------------------------------------------------------------------
import ppgplot
import string
from rea import ReaMessageHandler
#----------------------------------------------------------------------------------
#----- BoGLi Device Handler -------------------------------------------------------
#-------------------------------------------------------------------------

# attributes:
CurrentDev = 0
DevList = []
__MessHand = ReaMessageHandler.MessHand('DeviceHandler')

#--------------------------------------------------------------------------------
#----- public methods -----------------------------------------------------------
#-------------------------------------------------------------------------


def openDev(type='/XWINDOW'):
    """
    DES: open a device, return the device id
    INP: (string) type = pgplot device type
    """

    global CurrentDev, DevList

    if isinstance(type, str):
        CurrentDev = ppgplot.pgopen(type)
        ppgplot.pgask(0)
        DevList.append(CurrentDev)
    else:
        __MessHand.setMess(0, "openDev: string argument required ")
    __queryDev()
#-------------------------------------------------------------------------


def selectDev(devID=''):
    """
    DES: select an open device
    INP: (int) device ID
    """

    global CurrentDev, DevList

    if devID in DevList:
        ppgplot.pgslct(devID)             # select the deviceID
        CurrentDev = ppgplot.pgqid()        # store the actual pgplot deviceID
    else:
        __MessHand.error("this is not an open device: " + repr(devID))
    __queryDev()


#-------------------------------------------------------------------------
def closeDev(devID='current'):
    """
    DES: close selected device
    INP: (int) device ID, 'all','current' (default)
    """

    global CurrentDev, DevList

    if devID in ['current', 'curren', 'curre', 'curr', 'cur']:

        __MessHand.longinfo("closing the current device")
        ppgplot.pgclos()
        DevList.remove(CurrentDev)
        if not DevList == []:
            ppgplot.pgslct(DevList[0])
            CurrentDev = ppgplot.pgqid()        # store the actual deviceID
        else:
            CurrentDev = 0

    elif devID in ['all']:

        __MessHand.longinfo("closing all devices")
        for device in DevList:
            ppgplot.pgslct(device)
            ppgplot.pgclos()

        DevList = []
        CurrentDev = 0

    elif devID in DevList:

        __MessHand.longinfo("closing device " + repr(devID))
        ppgplot.pgslct(devID)
        ppgplot.pgclos()
        DevList.remove(devID)
        if devID == CurrentDev and DevList != []:
            ppgplot.pgslct(DevList[0])
            CurrentDev = ppgplot.pgqid()        # store the actual deviceID
        else:
            ppgplot.pgslct(CurrentDev)
    else:
        __MessHand.error("this is not an open device: " + repr(devID))
    __queryDev()

#-------------------------------------------------------------------------


def resizeDev():
    """
    DES: resize plot area after resizing window with mouse
    ABB: resize
    """
    __MessHand.info("resizing plot page")
    ppgplot.pgpage()


#-------------------------------------------------------------------------
def resizeDev():
    """
    Des: resize plot area after resizing window with mouse
    ABB: resize
    """
    __MessHand.info("resizing plot page")
    ppgplot.pgpage()


#--------------------------------------------------------------------------------
#----- private methods ----------------------------------------------------------
#-------------------------------------------------------------------------
def __queryDev():
    """
    DES: query open and current devices
    INP: (int) a = message weight
    """

    global CurrentDev, DevList

    DevList.sort()
    CurrentDev = ppgplot.pgqid()        # store the actual deviceID

    if CurrentDev:
        __MessHand.info("current device = " + repr(CurrentDev))
        if not CurrentDev in DevList:
            # for some reason a device is open and not in the DevList
            DevList.append(CurrentDev)
    else:
        __MessHand.warning("current device = none")

    if len(DevList) > 0:
        __MessHand.longinfo("open devices   = " + repr(DevList))
    else:
        __MessHand.warning("open devices   = none")


def __checkReopen():
    """
    DES: check if the current device is actually open and usable, or if it
         has to be reopenned (e.g. after a plot produced by apexCalibrator)
    """

    global CurrentDev, DevList

    ppgId = ppgplot.pgqid()
    if ppgId != CurrentDev:
        CurrentDev = ppgId
        ppgplot.pgask(0)
        DevList.append(CurrentDev)
